# -*- coding: utf-8 -*-
"""
This module offers timezone implementations subclassing the abstract
:py:`datetime.tzinfo` type. There are classes to handle tzfile format files
(usually are in :file:`/etc/localtime`, :file:`/usr/share/zoneinfo`, etc), TZ
environment string (in all known formats), given ranges (with help from
relative deltas), local machine timezone, fixed offset timezone, and UTC
timezone.
"""
import datetime
import struct
import time
import sys
import os

from six import string_types, PY3
from ._common import tzname_in_python2

try:
    from .win import tzwin, tzwinlocal
except ImportError:
    tzwin = tzwinlocal = None

relativedelta = None
parser = None
rrule = None

ZERO = datetime.timedelta(0)
EPOCHORDINAL = datetime.datetime.utcfromtimestamp(0).toordinal()

class tzutc(datetime.tzinfo):

    def utcoffset(self, dt):
        return ZERO

    def dst(self, dt):
        return ZERO

    @tzname_in_python2
    def tzname(self, dt):
        return "UTC"

    def __eq__(self, other):
        return (isinstance(other, tzutc) or
                (isinstance(other, tzoffset) and other._offset == ZERO))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "%s()" % self.__class__.__name__

    __reduce__ = object.__reduce__


class tzoffset(datetime.tzinfo):

    def __init__(self, name, offset):
        self._name = name
        self._offset = datetime.timedelta(seconds=offset)

    def utcoffset(self, dt):
        return self._offset

    def dst(self, dt):
        return ZERO

    @tzname_in_python2
    def tzname(self, dt):
        return self._name

    def __eq__(self, other):
        return (isinstance(other, tzoffset) and
                self._offset == other._offset)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "%s(%s, %s)" % (self.__class__.__name__,
                               repr(self._name),
                               self._offset.days*86400+self._offset.seconds)

    __reduce__ = object.__reduce__


class tzlocal(datetime.tzinfo):
    def __init__(self):
        self._std_offset = datetime.timedelta(seconds=-time.timezone)
        if time.daylight:
            self._dst_offset = datetime.timedelta(seconds=-time.altzone)
        else:
            self._dst_offset = self._std_offset

    def utcoffset(self, dt):
        if dt is None:
            return dt

        if self._isdst(dt):
            return self._dst_offset
        else:
            return self._std_offset

    def dst(self, dt):
        if self._isdst(dt):
            return self._dst_offset-self._std_offset
        else:
            return ZERO

    @tzname_in_python2
    def tzname(self, dt):
        return time.tzname[self._isdst(dt)]

    def _isdst(self, dt):
        # We can't use mktime here. It is unstable when deciding if
        # the hour near to a change is DST or not.
        #
        # timestamp = time.mktime((dt.year, dt.month, dt.day, dt.hour,
        #                         dt.minute, dt.second, dt.weekday(), 0, -1))
        # return time.localtime(timestamp).tm_isdst
        #
        # The code above yields the following result:
        #
        # >>> import tz, datetime
        # >>> t = tz.tzlocal()
        # >>> datetime.datetime(2003,2,15,23,tzinfo=t).tzname()
        # 'BRDT'
        # >>> datetime.datetime(2003,2,16,0,tzinfo=t).tzname()
        # 'BRST'
        # >>> datetime.datetime(2003,2,15,23,tzinfo=t).tzname()
        # 'BRST'
        # >>> datetime.datetime(2003,2,15,22,tzinfo=t).tzname()
        # 'BRDT'
        # >>> datetime.datetime(2003,2,15,23,tzinfo=t).tzname()
        # 'BRDT'
        #
        # Here is a more stable implementation:
        #
        timestamp = ((dt.toordinal() - EPOCHORDINAL) * 86400
                     + dt.hour * 3600
                     + dt.minute * 60
                     + dt.second)
        return time.localtime(timestamp+time.timezone).tm_isdst

    def __eq__(self, other):
        return (isinstance(other, tzlocal) and
                (self._std_offset == other._std_offset and
                 self._dst_offset == other._dst_offset))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "%s()" % self.__class__.__name__

    __reduce__ = object.__reduce__


class _ttinfo(object):
    __slots__ = ["offset", "delta", "isdst", "abbr", "isstd", "isgmt"]

    def __init__(self):
        for attr in self.__slots__:
            setattr(self, attr, None)

    def __repr__(self):
        l = []
        for attr in self.__slots__:
            value = getattr(self, attr)
            if value is not None:
                l.append("%s=%s" % (attr, repr(value)))
        return "%s(%s)" % (self.__class__.__name__, ", ".join(l))

    def __eq__(self, other):
        if not isinstance(other, _ttinfo):
            return False
        return (self.offset == other.offset and
                self.delta == other.delta and
                self.isdst == other.isdst and
                self.abbr == other.abbr and
                self.isstd == other.isstd and
                self.isgmt == other.isgmt)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getstate__(self):
        state = {}
        for name in self.__slots__:
            state[name] = getattr(self, name, None)
        return state

    def __setstate__(self, state):
        for name in self.__slots__:
            if name in state:
                setattr(self, name, state[name])


class tzfile(datetime.tzinfo):

    # http://www.twinsun.com/tz/tz-link.htm
    # ftp://ftp.iana.org/tz/tz*.tar.gz

    def __init__(self, fileobj, filename=None):
        file_opened_here = False
        if isinstance(fileobj, string_types):
            self._filename = fileobj
            fileobj = open(fileobj, 'rb')
            file_opened_here = True
        elif filename is not None:
            self._filename = filename
        elif hasattr(fileobj, "name"):
            self._filename = fileobj.name
        else:
            self._filename = repr(fileobj)

        # From tzfile(5):
        #
        # The time zone information files used by tzset(3)
        # begin with the magic characters "TZif" to identify
        # them as time zone information files, followed by
        # sixteen bytes reserved for future use, followed by
        # six four-byte values of type long, written in a
        # ``standard'' byte order (the high-order  byte
        # of the value is written first).
        try:
            if fileobj.read(4).decode() != "TZif":
                raise ValueError("magic not found")

            fileobj.read(16)

            (
                # The number of UTC/local indicators stored in the file.
                ttisgmtcnt,

                # The number of standard/wall indicators stored in the file.
                ttisstdcnt,

                # The number of leap seconds for which data is
                # stored in the file.
                leapcnt,

                # The number of "transition times" for which data
                # is stored in the file.
                timecnt,

                # The number of "local time types" for which data
                # is stored in the file (must not be zero).
                typecnt,

                # The  number  of  characters  of "time zone
                # abbreviation strings" stored in the file.
                charcnt,

            ) = struct.unpack(">6l", fileobj.read(24))

            # The above header is followed by tzh_timecnt four-byte
            # values  of  type long,  sorted  in ascending order.
            # These values are written in ``standard'' byte order.
            # Each is used as a transition time (as  returned  by
            # time(2)) at which the rules for computing local time
            # change.

            if timecnt:
                self._trans_list = struct.unpack(">%dl" % timecnt,
                                                 fileobj.read(timecnt*4))
            else:
                self._trans_list = []

            # Next come tzh_timecnt one-byte values of type unsigned
            # char; each one tells which of the different types of
            # ``local time'' types described in the file is associated
            # with the same-indexed transition time. These values
            # serve as indices into an array of ttinfo structures that
            # appears next in the file.

            if timecnt:
                self._trans_idx = struct.unpack(">%dB" % timecnt,
                                                fileobj.read(timecnt))
            else:
                self._trans_idx = []

            # Each ttinfo structure is written as a four-byte value
            # for tt_gmtoff  of  type long,  in  a  standard  byte
            # order, followed  by a one-byte value for tt_isdst
            # and a one-byte  value  for  tt_abbrind.   In  each
            # structure, tt_gmtoff  gives  the  number  of
            # seconds to be added to UTC, tt_isdst tells whether
            # tm_isdst should be set by  localtime(3),  and
            # tt_abbrind serves  as an index into the array of
            # time zone abbreviation characters that follow the
            # ttinfo structure(s) in the file.

            ttinfo = []

            for i in range(typecnt):
                ttinfo.append(struct.unpack(">lbb", fileobj.read(6)))

            abbr = fileobj.read(charcnt).decode()

            # Then there are tzh_leapcnt pairs of four-byte
            # values, written in  standard byte  order;  the
            # first  value  of  each pair gives the time (as
            # returned by time(2)) at which a leap second
            # occurs;  the  second  gives the  total  number of
            # leap seconds to be applied after the given time.
            # The pairs of values are sorted in ascending order
            # by time.

            # Not used, for now (but read anyway for correct file position)
            if leapcnt:
                leap = struct.unpack(">%dl" % (leapcnt*2),
                                     fileobj.read(leapcnt*8))

            # Then there are tzh_ttisstdcnt standard/wall
            # indicators, each stored as a one-byte value;
            # they tell whether the transition times associated
            # with local time types were specified as standard
            # time or wall clock time, and are used when
            # a time zone file is used in handling POSIX-style
            # time zone environment variables.

            if ttisstdcnt:
                isstd = struct.unpack(">%db" % ttisstdcnt,
                                      fileobj.read(ttisstdcnt))

            # Finally, there are tzh_ttisgmtcnt UTC/local
            # indicators, each stored as a one-byte value;
            # they tell whether the transition times associated
            # with local time types were specified as UTC or
            # local time, and are used when a time zone file
            # is used in handling POSIX-style time zone envi-
            # ronment variables.

            if ttisgmtcnt:
                isgmt = struct.unpack(">%db" % ttisgmtcnt,
                                      fileobj.read(ttisgmtcnt))

            # ** Everything has been read **
        finally:
            if file_opened_here:
                fileobj.close()

        # Build ttinfo list
        self._ttinfo_list = []
        for i in range(typecnt):
            gmtoff, isdst, abbrind = ttinfo[i]
            # Round to full-minutes if that's not the case. Python's
            # datetime doesn't accept sub-minute timezones. Check
            # http://python.org/sf/1447945 for some information.
            gmtoff = (gmtoff+30)//60*60
            tti = _ttinfo()
            tti.offset = gmtoff
            tti.delta = datetime.timedelta(seconds=gmtoff)
            tti.isdst = isdst
            tti.abbr = abbr[abbrind:abbr.find('\x00', abbrind)]
            tti.isstd = (ttisstdcnt > i and isstd[i] != 0)
            tti.isgmt = (ttisgmtcnt > i and isgmt[i] != 0)
            self._ttinfo_list.append(tti)

        # Replace ttinfo indexes for ttinfo objects.
        trans_idx = []
        for idx in self._trans_idx:
            trans_idx.append(self._ttinfo_list[idx])
        self._trans_idx = tuple(trans_idx)

        # Set standard, dst, and before ttinfos. before will be
        # used when a given time is before any transitions,
        # and will be set to the first non-dst ttinfo, or to
        # the first dst, if all of them are dst.
        self._ttinfo_std = None
        self._ttinfo_dst = None
        self._ttinfo_before = None
        if self._ttinfo_list:
            if not self._trans_list:
                self._ttinfo_std = self._ttinfo_first = self._ttinfo_list[0]
            else:
                for i in range(timecnt-1, -1, -1):
                    tti = self._trans_idx[i]
                    if not self._ttinfo_std and not tti.isdst:
                        self._ttinfo_std = tti
                    elif not self._ttinfo_dst and tti.isdst:
                        self._ttinfo_dst = tti
                    if self._ttinfo_std and self._ttinfo_dst:
                        break
                else:
                    if self._ttinfo_dst and not self._ttinfo_std:
                        self._ttinfo_std = self._ttinfo_dst

                for tti in self._ttinfo_list:
                    if not tti.isdst:
                        self._ttinfo_before = tti
                        break
                else:
                    self._ttinfo_before = self._ttinfo_list[0]

        # Now fix transition times to become relative to wall time.
        #
        # I'm not sure about this. In my tests, the tz source file
        # is setup to wall time, and in the binary file isstd and
        # isgmt are off, so it should be in wall time. OTOH, it's
        # always in gmt time. Let me know if you have comments
        # about this.
        laststdoffset = 0
        self._trans_list = list(self._trans_list)
        for i in range(len(self._trans_list)):
            tti = self._trans_idx[i]
            if not tti.isdst:
                # This is std time.
                self._trans_list[i] += tti.offset
                laststdoffset = tti.offset
            else:
                # This is dst time. Convert to std.
                self._trans_list[i] += laststdoffset
        self._trans_list = tuple(self._trans_list)

    def _find_ttinfo(self, dt, laststd=0):
        timestamp = ((dt.toordinal() - EPOCHORDINAL) * 86400
                     + dt.hour * 3600
                     + dt.minute * 60
                     + dt.second)
        idx = 0
        for trans in self._trans_list:
            if timestamp < trans:
                break
            idx += 1
        else:
            return self._ttinfo_std
        if idx == 0:
            return self._ttinfo_before
        if laststd:
            while idx > 0:
                tti = self._trans_idx[idx-1]
                if not tti.isdst:
                    return tti
                idx -= 1
            else:
                return self._ttinfo_std
        else:
            return self._trans_idx[idx-1]

    def utcoffset(self, dt):
        if dt is None:
            return None

        if not self._ttinfo_std:
            return ZERO
        return self._find_ttinfo(dt).delta

    def dst(self, dt):
        if not self._ttinfo_dst:
            return ZERO
        tti = self._find_ttinfo(dt)
        if not tti.isdst:
            return ZERO

        # The documentation says that utcoffset()-dst() must
        # be constant for every dt.
        return tti.delta-self._find_ttinfo(dt, laststd=1).delta

        # An alternative for that would be:
        #
        # return self._ttinfo_dst.offset-self._ttinfo_std.offset
        #
        # However, this class stores historical changes in the
        # dst offset, so I belive that this wouldn't be the right
        # way to implement this.

    @tzname_in_python2
    def tzname(self, dt):
        if not self._ttinfo_std:
            return None
        return self._find_ttinfo(dt).abbr

    def __eq__(self, other):
        if not isinstance(other, tzfile):
            return False
        return (self._trans_list == other._trans_list and
                self._trans_idx == other._trans_idx and
                self._ttinfo_list == other._ttinfo_list)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, repr(self._filename))

    def __reduce__(self):
        if not os.path.isfile(self._filename):
            raise ValueError("Unpickable %s class" % self.__class__.__name__)
        return (self.__class__, (self._filename,))


class tzrange(datetime.tzinfo):
    def __init__(self, stdabbr, stdoffset=None,
                 dstabbr=None, dstoffset=None,
                 start=None, end=None):
        global relativedelta
        if not relativedelta:
            from dateutil import relativedelta
        self._std_abbr = stdabbr
        self._dst_abbr = dstabbr
        if stdoffset is not None:
            self._std_offset = datetime.timedelta(seconds=stdoffset)
        else:
            self._std_offset = ZERO
        if dstoffset is not None:
            self._dst_offset = datetime.timedelta(seconds=dstoffset)
        elif dstabbr and stdoffset is not None:
            self._dst_offset = self._std_offset+datetime.timedelta(hours=+1)
        else:
            self._dst_offset = ZERO
        if dstabbr and start is None:
            self._start_delta = relativedelta.relativedelta(
                hours=+2, month=4, day=1, weekday=relativedelta.SU(+1))
        else:
            self._start_delta = start
        if dstabbr and end is None:
            self._end_delta = relativedelta.relativedelta(
                hours=+1, month=10, day=31, weekday=relativedelta.SU(-1))
        else:
            self._end_delta = end

    def utcoffset(self, dt):
        if dt is None:
            return None

        if self._isdst(dt):
            return self._dst_offset
        else:
            return self._std_offset

    def dst(self, dt):
        if self._isdst(dt):
            return self._dst_offset-self._std_offset
        else:
            return ZERO

    @tzname_in_python2
    def tzname(self, dt):
        if self._isdst(dt):
            return self._dst_abbr
        else:
            return self._std_abbr

    def _isdst(self, dt):
        if not self._start_delta:
            return False
        year = datetime.datetime(dt.year, 1, 1)
        start = year+self._start_delta
        end = year+self._end_delta
        dt = dt.replace(tzinfo=None)
        if start < end:
            return dt >= start and dt < end
        else:
            return dt >= start or dt < end

    def __eq__(self, other):
        if not isinstance(other, tzrange):
            return False
        return (self._std_abbr == other._std_abbr and
                self._dst_abbr == other._dst_abbr and
                self._std_offset == other._std_offset and
                self._dst_offset == other._dst_offset and
                self._start_delta == other._start_delta and
                self._end_delta == other._end_delta)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "%s(...)" % self.__class__.__name__

    __reduce__ = object.__reduce__


class tzstr(tzrange):

    def __init__(self, s):
        global parser
        if not parser:
            from dateutil import parser
        self._s = s

        res = parser._parsetz(s)
        if res is None:
            raise ValueError("unknown string format")

        # Here we break the compatibility with the TZ variable handling.
        # GMT-3 actually *means* the timezone -3.
        if res.stdabbr in ("GMT", "UTC"):
            res.stdoffset *= -1

        # We must initialize it first, since _delta() needs
        # _std_offset and _dst_offset set. Use False in start/end
        # to avoid building it two times.
        tzrange.__init__(self, res.stdabbr, res.stdoffset,
                         res.dstabbr, res.dstoffset,
                         start=False, end=False)

        if not res.dstabbr:
            self._start_delta = None
            self._end_delta = None
        else:
            self._start_delta = self._delta(res.start)
            if self._start_delta:
                self._end_delta = self._delta(res.end, isend=1)

    def _delta(self, x, isend=0):
        kwargs = {}
        if x.month is not None:
            kwargs["month"] = x.month
            if x.weekday is not None:
                kwargs["weekday"] = relativedelta.weekday(x.weekday, x.week)
                if x.week > 0:
                    kwargs["day"] = 1
                else:
                    kwargs["day"] = 31
            elif x.day:
                kwargs["day"] = x.day
        elif x.yday is not None:
            kwargs["yearday"] = x.yday
        elif x.jyday is not None:
            kwargs["nlyearday"] = x.jyday
        if not kwargs:
            # Default is to start on first sunday of april, and end
            # on last sunday of october.
            if not isend:
                kwargs["month"] = 4
                kwargs["day"] = 1
                kwargs["weekday"] = relativedelta.SU(+1)
            else:
                kwargs["month"] = 10
                kwargs["day"] = 31
                kwargs["weekday"] = relativedelta.SU(-1)
        if x.time is not None:
            kwargs["seconds"] = x.time
        else:
            # Default is 2AM.
            kwargs["seconds"] = 7200
        if isend:
            # Convert to standard time, to follow the documented way
            # of working with the extra hour. See the documentation
            # of the tzinfo class.
            delta = self._dst_offset-self._std_offset
            kwargs["seconds"] -= delta.seconds+delta.days*86400
        return relativedelta.relativedelta(**kwargs)

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, repr(self._s))


class _tzicalvtzcomp(object):
    def __init__(self, tzoffsetfrom, tzoffsetto, isdst,
                 tzname=None, rrule=None):
        self.tzoffsetfrom = datetime.timedelta(seconds=tzoffsetfrom)
        self.tzoffsetto = datetime.timedelta(seconds=tzoffsetto)
        self.tzoffsetdiff = self.tzoffsetto-self.tzoffsetfrom
        self.isdst = isdst
        self.tzname = tzname
        self.rrule = rrule


class _tzicalvtz(datetime.tzinfo):
    def __init__(self, tzid, comps=[]):
        self._tzid = tzid
        self._comps = comps
        self._cachedate = []
        self._cachecomp = []

    def _find_comp(self, dt):
        if len(self._comps) == 1:
            return self._comps[0]
        dt = dt.replace(tzinfo=None)
        try:
            return self._cachecomp[self._cachedate.index(dt)]
        except ValueError:
            pass
        lastcomp = None
        lastcompdt = None
        for comp in self._comps:
            if not comp.isdst:
                # Handle the extra hour in DST -> STD
                compdt = comp.rrule.before(dt-comp.tzoffsetdiff, inc=True)
            else:
                compdt = comp.rrule.before(dt, inc=True)
            if compdt and (not lastcompdt or lastcompdt < compdt):
                lastcompdt = compdt
                lastcomp = comp
        if not lastcomp:
            # RFC says nothing about what to do when a given
            # time is before the first onset date. We'll look for the
            # first standard component, or the first component, if
            # none is found.
            for comp in self._comps:
                if not comp.isdst:
                    lastcomp = comp
                    break
            else:
                lastcomp = comp[0]
        self._cachedate.insert(0, dt)
        self._cachecomp.insert(0, lastcomp)
        if len(self._cachedate) > 10:
            self._cachedate.pop()
            self._cachecomp.pop()
        return lastcomp

    def utcoffset(self, dt):
        if dt is None:
            return None

        return self._find_comp(dt).tzoffsetto

    def dst(self, dt):
        comp = self._find_comp(dt)
        if comp.isdst:
            return comp.tzoffsetdiff
        else:
            return ZERO

    @tzname_in_python2
    def tzname(self, dt):
        return self._find_comp(dt).tzname

    def __repr__(self):
        return "<tzicalvtz %s>" % repr(self._tzid)

    __reduce__ = object.__reduce__


class tzical(object):
    def __init__(self, fileobj):
        global rrule
        if not rrule:
            from dateutil import rrule

        if isinstance(fileobj, string_types):
            self._s = fileobj
            # ical should be encoded in UTF-8 with CRLF
            fileobj = open(fileobj, 'r')
        elif hasattr(fileobj, "name"):
            self._s = fileobj.name
        else:
            self._s = repr(fileobj)

        self._vtz = {}

        self._parse_rfc(fileobj.read())

    def keys(self):
        return list(self._vtz.keys())

    def get(self, tzid=None):
        if tzid is None:
            keys = list(self._vtz.keys())
            if len(keys) == 0:
                raise ValueError("no timezones defined")
            elif len(keys) > 1:
                raise ValueError("more than one timezone available")
            tzid = keys[0]
        return self._vtz.get(tzid)

    def _parse_offset(self, s):
        s = s.strip()
        if not s:
            raise ValueError("empty offset")
        if s[0] in ('+', '-'):
            signal = (-1, +1)[s[0] == '+']
            s = s[1:]
        else:
            signal = +1
        if len(s) == 4:
            return (int(s[:2])*3600+int(s[2:])*60)*signal
        elif len(s) == 6:
            return (int(s[:2])*3600+int(s[2:4])*60+int(s[4:]))*signal
        else:
            raise ValueError("invalid offset: "+s)

    def _parse_rfc(self, s):
        lines = s.splitlines()
        if not lines:
            raise ValueError("empty string")

        # Unfold
        i = 0
        while i < len(lines):
            line = lines[i].rstrip()
            if not line:
                del lines[i]
            elif i > 0 and line[0] == " ":
                lines[i-1] += line[1:]
                del lines[i]
            else:
                i += 1

        tzid = None
        comps = []
        invtz = False
        comptype = None
        for line in lines:
            if not line:
                continue
            name, value = line.split(':', 1)
            parms = name.split(';')
            if not parms:
                raise ValueError("empty property name")
            name = parms[0].upper()
            parms = parms[1:]
            if invtz:
                if name == "BEGIN":
                    if value in ("STANDARD", "DAYLIGHT"):
                        # Process component
                        pass
                    else:
                        raise ValueError("unknown component: "+value)
                    comptype = value
                    founddtstart = False
                    tzoffsetfrom = None
                    tzoffsetto = None
                    rrulelines = []
                    tzname = None
                elif name == "END":
                    if value == "VTIMEZONE":
                        if comptype:
                            raise ValueError("component not closed: "+comptype)
                        if not tzid:
                            raise ValueError("mandatory TZID not found")
                        if not comps:
                            raise ValueError(
                                "at least one component is needed")
                        # Process vtimezone
                        self._vtz[tzid] = _tzicalvtz(tzid, comps)
                        invtz = False
                    elif value == comptype:
                        if not founddtstart:
                            raise ValueError("mandatory DTSTART not found")
                        if tzoffsetfrom is None:
                            raise ValueError(
                                "mandatory TZOFFSETFROM not found")
                        if tzoffsetto is None:
                            raise ValueError(
                                "mandatory TZOFFSETFROM not found")
                        # Process component
                        rr = None
                        if rrulelines:
                            rr = rrule.rrulestr("\n".join(rrulelines),
                                                compatible=True,
                                                ignoretz=True,
                                                cache=True)
                        comp = _tzicalvtzcomp(tzoffsetfrom, tzoffsetto,
                                              (comptype == "DAYLIGHT"),
                                              tzname, rr)
                        comps.append(comp)
                        comptype = None
                    else:
                        raise ValueError("invalid component end: "+value)
                elif comptype:
                    if name == "DTSTART":
                        rrulelines.append(line)
                        founddtstart = True
                    elif name in ("RRULE", "RDATE", "EXRULE", "EXDATE"):
                        rrulelines.append(line)
                    elif name == "TZOFFSETFROM":
                        if parms:
                            raise ValueError(
                                "unsupported %s parm: %s " % (name, parms[0]))
                        tzoffsetfrom = self._parse_offset(value)
                    elif name == "TZOFFSETTO":
                        if parms:
                            raise ValueError(
                                "unsupported TZOFFSETTO parm: "+parms[0])
                        tzoffsetto = self._parse_offset(value)
                    elif name == "TZNAME":
                        if parms:
                            raise ValueError(
                                "unsupported TZNAME parm: "+parms[0])
                        tzname = value
                    elif name == "COMMENT":
                        pass
                    else:
                        raise ValueError("unsupported property: "+name)
                else:
                    if name == "TZID":
                        if parms:
                            raise ValueError(
                                "unsupported TZID parm: "+parms[0])
                        tzid = value
                    elif name in ("TZURL", "LAST-MODIFIED", "COMMENT"):
                        pass
                    else:
                        raise ValueError("unsupported property: "+name)
            elif name == "BEGIN" and value == "VTIMEZONE":
                tzid = None
                comps = []
                invtz = True

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, repr(self._s))

if sys.platform != "win32":
    TZFILES = ["/etc/localtime", "localtime"]
    TZPATHS = ["/usr/share/zoneinfo", "/usr/lib/zoneinfo", "/etc/zoneinfo"]
else:
    TZFILES = []
    TZPATHS = []


def gettz(name=None):
    tz = None
    if not name:
        try:
            name = os.environ["TZ"]
        except KeyError:
            pass
    if name is None or name == ":":
        for filepath in TZFILES:
            if not os.path.isabs(filepath):
                filename = filepath
                for path in TZPATHS:
                    filepath = os.path.join(path, filename)
                    if os.path.isfile(filepath):
                        break
                else:
                    continue
            if os.path.isfile(filepath):
                try:
                    tz = tzfile(filepath)
                    break
                except (IOError, OSError, ValueError):
                    pass
        else:
            tz = tzlocal()
    else:
        if name.startswith(":"):
            name = name[:-1]
        if os.path.isabs(name):
            if os.path.isfile(name):
                tz = tzfile(name)
            else:
                tz = None
        else:
            for path in TZPATHS:
                filepath = os.path.join(path, name)
                if not os.path.isfile(filepath):
                    filepath = filepath.replace(' ', '_')
                    if not os.path.isfile(filepath):
                        continue
                try:
                    tz = tzfile(filepath)
                    break
                except (IOError, OSError, ValueError):
                    pass
            else:
                tz = None
                if tzwin is not None:
                    try:
                        tz = tzwin(name)
                    except WindowsError:
                        tz = None
                if not tz:
                    from dateutil.zoneinfo import gettz
                    tz = gettz(name)
                if not tz:
                    for c in name:
                        # name must have at least one offset to be a tzstr
                        if c in "0123456789":
                            try:
                                tz = tzstr(name)
                            except ValueError:
                                pass
                            break
                    else:
                        if name in ("GMT", "UTC"):
                            tz = tzutc()
                        elif name in time.tzname:
                            tz = tzlocal()
    return tz

# vim:ts=4:sw=4:et



import string
from pprint import pformat
from collections import OrderedDict
from datetime import datetime, date
from decimal import Decimal
import itertools
import os
base_path = os.path.dirname(os.path.abspath(__file__))

try:
    from UserDict import DictMixin
except ImportError:
    from collections import MutableMapping as DictMixin

try:
    text = unicode
except NameError:
    text = str 


# Use the same json implementation as itsdangerous because Flask does that
# and we depend on it
try:                                                                               
    from itsdangerous import simplejson as _json                                
except ImportError:                                                             
    from itsdangerous import json as _json                                      


def convert_column_name(column_name):
    #TODO: only A-z0-9_ allowd sanitize with regexp, throw exception
    column_name = column_name.lower()
    column_name = column_name.replace(' ','_')
    return column_name.replace('-','_')

def to_camel_case(snake_str):
    components = snake_str.split('_')
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + ''.join(x.title() for x in components[1:])

def to_kebab_case(column_name):
    return column_name.replace('_','-')


class ColumnNameGenerator(object):

    def __init__(self):
        self.header_memory = []
        self.collisions = {}
        self.generic_name_counter = 0    
        self.letters_and_digits = string.ascii_lowercase + string.digits

    def next_name(self, orignal_name):

        buf = ''

        if len(orignal_name) > 0:

            for i, char in enumerate(orignal_name.lower()):

                if i == 0 and char in string.digits:
                    buf += '_'

                elif char not in self.letters_and_digits:
                    buf += '_'

                else:
                    buf += char

            # check for collisions
            if buf in self.header_memory:
                suffix = self.collisions.get(buf)

                if suffix is not None:
                    self.collisions[buf] = suffix+1 
                    buf = buf + '_' +  str(suffix+1)

                else:
                    buf += '_1'

        else:
            self.generic_name_counter += 1
            buf = 'column_%i' % self.generic_name_counter

        self.header_memory.append(buf)

        return buf


def coalesce(*values):
    for v in values:
        if v is not None:
            return v  


def chainlist(iter_of_lists):
    return list(itertools.chain(*iter_of_lists))


class TupleMeta(object):

    def __init__(self, input_data=None, select_vector=None, **kw_args):
        buf = {}

        if input_data is None:
            input_data = kw_args

        elif isinstance(input_data, list):
            input_data = dict(input_data)

        elif not (isinstance(input_data, dict) or isinstance(input_data, TupleMeta)):
            raise ValueError('input must be either of type dict or list')

        if select_vector is not None and len(select_vector) != 0:

            if len(select_vector) > len(self.__slots__):
                raise ValueError('Source select vector is too long')
        else:
            select_vector = self.__slots__

        for k,v in input_data.items():
            buf[convert_column_name(k)] = v 
        
        for i, slot_key in enumerate(self.__slots__):

            if i < len(select_vector):
                setattr(self, slot_key, buf.get(select_vector[i]))
            else:
                setattr(self, slot_key, buf.get(slot_key))

    def __getitem__(self, k):
        if isinstance(k, int):
            return getattr(self,self.__slots__[k])

        if k in self.__slots__:
            return getattr(self, k)

    def __contains__(self, value):
        return value in self.__slots__

    def __eq__(self, value):
        if len(self.__slots__) == 1:
            stored_value = self[0]

            if isinstance(value, type(stored_value)) \
                or (isinstance(stored_value, text) and isinstance(value, str)) \
                or (isinstance(stored_value, str) and isinstance(value, text)):

                return stored_value == value
       
        for i, k in enumerate(self.__slots__):
            if getattr(self, k) != value[i]:
                return False

        return True

    def __ne__(self, value):

        for k in self.__slots__:
            if getattr(self, k) != value[k]:
                return True

        return False

    def __repr__(self):
        d = self.items()
        return 'Tuple %s' % d.__repr__()

    def __hash__(self):

        def hash_value(v):
            
            if isinstance(v, dict):
                return frozenset(v.items())

            return v

        if len(self.__slots__) == 1:
            return hash(hash_value(self.values()[0]))

        return hash(tuple([hash_value(v) for v in self.values()]))

    def __len__(self):
        return len(self.__slots__)

    def get(self, k, default=None):
        if k in self.__slots__:

            try:
                return getattr(self, k)

            except AttributeError:
                return default

        return default

    def keys(self):
        return self.__slots__

    def items(self):
        result = []
        for k in self.__slots__:
            result.append((k, getattr(self, k)))

        return result

    def values(self):
        result = []
        for k in self.__slots__:
            if hasattr(self,k):
                result.append(getattr(self, k))
            else:
                result.append(getattr(self, None))

        return result

    def select(self, keys):
        result = {}

        for k in keys:
            if k in self.__slots__:
                result[k] = getattr(self, k)

        return result

    def __dict__(self):
        return self.asdict()

    def asdict(self):
        result = {}

        for k in self.__slots__:
            result[k] = getattr(self, k) 

        return result

    def asjson(self):
        return _json.dumps(self.asdict(), cls=TupleJSONEncoder)


def build_tuple_type(*columns):

    if len(columns) == 1 and isinstance(columns[0], list):
        columns = columns[0]

    columns = [convert_column_name(c) for c in columns]

    class Tuple(TupleMeta): 
        __slots__ = columns

    return Tuple


class TupleJSONEncoder(_json.JSONEncoder):
    
    def default(self, obj):

        if isinstance(obj, datetime):
            return obj.isoformat()

        if isinstance(obj, date):
            return obj.isoformat()

        if isinstance(obj, Decimal):
            return str(obj)

        asdict_op = getattr(obj, "asdict", None)
        if callable(asdict_op):
            return obj.asdict()

        else:
            return _json.JSONEncoder.default(self, obj)


class BaseItem(object):
    """Base class for all scraped items."""
    pass


class Field(dict):
    """Container of field metadata"""
    pass


class ItemMeta(type):

    def __new__(mcs, class_name, bases, attrs):
        fields = OrderedDict()
        new_attrs = OrderedDict()

        for n, v in attrs.iteritems():
            if isinstance(v, Field):
                fields[n] = v
            else:
                new_attrs[n] = v

        cls = super(ItemMeta, mcs).__new__(mcs, class_name, bases, new_attrs)
        cls.fields = cls.fields.copy()
        cls.fields.update(fields)
        return cls


class DictItem(DictMixin, BaseItem):

    fields = OrderedDict()

    def __init__(self, *args, **kwargs):
        self._values = OrderedDict()
        if args or kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).iteritems():
                self[k] = v

    def __getitem__(self, key):
        if key in self.fields:
            return self._values.get(key)

        else:
            raise KeyError("%s does not support field: %s" %
                (self.__class__.__name__, key))

    def __setitem__(self, key, value):
        if key in self.fields:
            self._values[key] = value
        else:
            raise KeyError("%s does not support field: %s" %
                (self.__class__.__name__, key))

    def __delitem__(self, key):
        del self._values[key]

    def __getattr__(self, name):
        if name in self.fields:
            raise AttributeError("Use item[%r] to get field value" % name)
        raise AttributeError(name)

    def __setattr__(self, name, value):
        if not name.startswith('_'):
            raise AttributeError("Use item[%r] = %r to set field value" %
                (name, value))
        super(DictItem, self).__setattr__(name, value)

    def keys(self):
        return self._values.keys()

    def __repr__(self):
        return pformat(dict(self))

    def copy(self):
        return self.__class__(self)

    def __len__(self):
        return len(self.fields)

    def __iter__(self):
        return iter(self.fields.keys())
    
    def is_empty(self):
        return len(self._values) == 0

    def load(self, source):
        getter = None

        def get_item_value(c, keys):

            for k in keys:
                v = source.get(k)

                if v is not None:
                    return v

            return None

        if isinstance(source, list):
            def faa(c,k):
                if c >= len(source):
                    return None

                return source[c]

            getter = faa
            
            #getter = lambda c,k: source[c]

        elif isinstance(source, dict):
            getter = get_item_value

        else:
            asdict_op = getattr(source, "asdict", None)
            if callable(asdict_op):
                source = source.asdict()
                getter = get_item_value
            else:
                raise ValueError('Input not recognized. It must be list,dict or have asdict method')

        for c, key in enumerate(self.fields.keys()):
            alt_keys = self.fields[key].get('source_names', [])

            if self.fields[key].get('from_camel_case', False):
                alt_keys = [to_camel_case(key)]

            if self.fields[key].get('from_kebab_case', False):
                alt_keys = [to_kebab_case(key)]
                
            value = getter(c, [key] + alt_keys)

            if value is None:
                self[key] = None

            elif (isinstance(value, str) or isinstance(value, text)) and value == '' and self.fields[key].get('psql_type') in ('text', 'bytea'):
                self[key] = None

            elif (isinstance(value, str) or isinstance(value, text)) and value == '' and self.fields[key].get('psql_type') in ('int', 'bigint', 'numeric'):
                self[key] = None
            
            elif isinstance(value, str):
                self[key] = value

            elif isinstance(value, text):
                self[key] = value.encode('utf-8', 'ignore')
            
            elif isinstance(value, list):
                self[key] = _json.dumps(value, cls=TupleJSONEncoder)

            elif isinstance(value, dict):
                self[key] = _json.dumps(value, cls=TupleJSONEncoder)
            
            else:
                self[key] = value


    def select(self, keys):
        result = []

        for key in keys:
            if key in self.keys():
                result.append(self[key])

        return result

    def add_field(self, name, **kwargs):
        field = Field(**kwargs)
        self.fields[name] = field


class Item(DictItem):

    __metaclass__ = ItemMeta


def read_batch(iterator, batch_size=None):
    
    if batch_size is None:
        batch_size = 10000

    result = []

    for i in range(batch_size):
        
        try:
            result.append(next(iterator))

        except StopIteration:
            return result

    return result 


def remove_null_rows(batch, primary_key):

    def ff(item):
        for k in primary_key:
            if item[k] is None:
                return False

        empty_count = 0

        for c in item.values():
            if c is None:
                empty_count +=1

        if empty_count == len(item):
            return False

        #print (item, len(item), empty_count, item.is_empty())
        return True

    return list(filter(ff,batch))


class seekable(object):
    
    def __init__(self, iterable):
        self.pos = 0
        self.buf = []
        self.buf_iterator = iter(self.buf)
        self.disabled = False

    def __iter__(self):
        return self

    def __next__(self):
        self.disabled = True

        if self.pos < len(self.buf):
            b=self.buf[self.pos]
            self.pos+=1
            return b
        else:
            return next(self.iterator)

    def __getitem__(self, index):

        # getting an item is still posible but we are rather raising error
        # to prevent incorrect usage
        if not self.disabled:

            if index < len(self.buf):
                return self.buf[index]
            else:
                self.seek(index)
                return self.buf[index]

        else:
            raise ValueError('Seeking disabled iteration already started')

    def seek(self, index):

        if not self.disabled:

            for i in range(index-(len(self.buf)-1)):
                self.buf.append(next(self.iterator))
                    
        else:
            raise ValueError('Seeking disabled iteration already started')

    def is_empty(self):
        try:
            self.seek(1)
            return False

        except StopIteration:
            return True
     
    def has_one_element(self):
        test_first = False
        test_second = False

        try:
            self.seek(1)
            test_first = True

        except StopIteration:
            test_first = False

        try:
            self.seek(2)
            test_first = False

        except StopIteration:
            test_first = True

        return test_first and not test_second

    def get_buffer_iterator(self):
        return iter(self.buf)


__query_paths = ['./', './queries', './templates', os.path.join(base_path, 'queries')]

def find_query_template(template_name):

    if not template_name.endswith('.sql'):
        template_name += '.sql'

    template_name = template_name.lstrip(u'\ufdd0:')

    filename_patterns = [template_name,
                         template_name.replace('-', '_'),
                         template_name.replace('_', '-')]   

    path_pattern = itertools.product(__query_paths, filename_patterns)
    for t in path_pattern:
        file_path = os.path.join(t[0], t[1])
        if os.path.isfile(file_path):

            with open(file_path) as query_file:
                query_tpl = query_file.read()

            return query_tpl




# -*- coding: utf-8 -*-
import string
from decimal import Decimal
import csv
import re
import unidecode
from collections import OrderedDict
from witchcraft.combinators import execute, query, template
from witchcraft.dateutil.parser import parse as dateutil_parse


prefix_dict = {
    'pgsql': 'psql',
    'mysql': 'mysql',
    'oracle': 'oracle',
    'mssql': 'mssql', 
}
     

def find_keys(dps):
    keys = set()

    for i in dps:
        keys |= set(i.keys())

    return keys


#TODO: get rid of side effect
def remove_metadata(data_points):
    
    for item in data_points:
        if '_updated_at' in item:
            del item['_updated_at']

        if '_created_at' in item:
            del item['_created_at']


def create_table(connection, schema_name, table_name, fields, primary_keys, version_column=None):

    primary_keys = list(primary_keys)


    columns = list(fields.items())

    if version_column is not None:
        primary_keys.append(version_column)
        columns.append((version_column, {'psql_type': 'bigint'}))

    prefix = prefix_dict.get(connection.database_type)
    execute(connection, template('%s_create_table' % prefix, 
                            dict(columns=columns,
                                 schema_name=schema_name,
                                 table_name=table_name,
                                 primary_keys=primary_keys)))

# TODO: implement column upgrade/add as we go through data points To make this 
# possible DataPoint have to implement simple way of checking if fields sqltype changed.
# 

def prepare_table(connection, schema_name, table_name, data_points, primary_keys, version_column=None):
    fields = OrderedDict([])

    for dp in data_points:
        fields.update(dp.fields)

    prefix = prefix_dict.get(connection.database_type)

    required_columns = find_keys(data_points)
    discovery_table_name = table_name

    if connection.database_type == 'oracle':
        discovery_table_name = table_name.upper()

    result = query(connection, template('%s_discover_columns' % prefix,
                                     dict(schema_name=schema_name,
                                          table_name=discovery_table_name)))

    discovered_columns = set(map(lambda r: r.column_name.lower(), result))
    
    if len(discovered_columns) == 0:
        create_table(connection, schema_name, table_name, fields, primary_keys)

        result = query(connection, template('%s_discover_columns' % prefix,
                                     dict(schema_name=schema_name,
                                          table_name=discovery_table_name)))

        discovered_columns = set(map(lambda r: r.column_name.lower(), result))

    else:
        discovered_pkeys = filter(lambda r: r.is_pkey, result)
        discovered_pkeys = list(map(lambda r: r.column_name.lower(), discovered_pkeys))

        if version_column is not None:
            discovered_pkeys.remove(version_column)

        if len(primary_keys) != 0:

            if set(primary_keys) != set(discovered_pkeys):
                raise ValueError('Primary keys in destination table are not matching with defined primary keys %s %s' % (set(primary_keys),set(discovered_pkeys)))
            primary_keys = discovered_pkeys

    missing_columns = required_columns - discovered_columns

    for column_name in missing_columns:
        column_type = fields.get(column_name)['%s_type' % prefix]

        execute(connection, template('%s_add_column' % prefix,
                                dict(schema_name=schema_name,
                                     table_name=table_name,
                                     column_name=column_name,
                                     column_type=column_type)))

    if version_column is not None and version_column not in discovered_columns:
        execute(connection, template('%s_add_column' % prefix,
                                dict(schema_name=schema_name,
                                     table_name=table_name,
                                     column_name=version_column,
                                     column_type='bigint')))

    return primary_keys


def upsert_data(connection, schema_name, table_name, data_points, primary_keys):

    prefix = prefix_dict.get(connection.database_type)
    column_names = list(find_keys(data_points))

    connection.begin()

    execute(connection, template('%s_upsert_load' % prefix,
                        dict(schema_name=schema_name,
                             table_name=table_name,
                             column_names=column_names,
                             columns=data_points[0].fields.items(),
                             data_points=data_points,
                             primary_keys=primary_keys),
                        connection.database_type))

    updated = execute(connection, template('%s_upsert_update' % prefix,
                        dict(schema_name=schema_name,
                             table_name=table_name,
                             column_names=column_names,
                             columns=data_points[0].fields.items(),
                             data_points=data_points,
                             primary_keys=primary_keys),
                        connection.database_type))

    inserted = execute(connection, template('%s_upsert_insert' % prefix,
                        dict(schema_name=schema_name,
                             table_name=table_name,
                             column_names=column_names,
                             columns=data_points[0].fields.items(),
                             data_points=data_points,
                             primary_keys=primary_keys),
                        connection.database_type))
    connection.commit()

    return (inserted, updated)


def insert_data(connection, schema_name, table_name, data_points):
    prefix = prefix_dict.get(connection.database_type)
    column_names = list(find_keys(data_points))

    inserted = query(connection, 
                   template('%s_insert' % prefix,
                            dict(schema_name=schema_name,
                                 table_name=table_name,
                                 column_names=column_names,
                                 columns=data_points[0].fields.items(),
                                 data_points=data_points),
                            connection.database_type))

    connection.commit()
    return inserted[0].count


def delete_data(connection, schema_name, table_name):
    prefix = prefix_dict.get(connection.database_type)

    result = query(connection, template('%s_discover_columns' % prefix,
                                        dict(schema_name=schema_name,
                                             table_name=table_name)))

    if len(result) == 0:
        return 0


    #TODO: use SqlSession.delete directly?
    deleted_count = execute(connection,
                            template('%s_delete' % prefix,
                                     dict(schema_name=schema_name,
                                          table_name=table_name),
                                          connection.database_type))

    return deleted_count
   

def discover_columns(connection, schema_name, table_name):
    prefix = prefix_dict.get(connection.database_type)
    result = query(connection, template('%s_discover_columns' % prefix,
                                 dict(schema_name=schema_name,
                                      table_name=table_name)))

    return result


def get_max_version(connection, schema_name, table_name, version_column='version'):
    prefix = prefix_dict.get(connection.database_type)

    result = query(connection, template('%s_discover_columns' % prefix,
                                        dict(schema_name=schema_name,
                                             table_name=table_name)))

    if len(result) == 0:
        return 1

    tpl = template('%s_max_version' % prefix,
                     dict(schema_name=schema_name,
                          table_name=table_name,
                          version_column=version_column),
                     connection.database_type)

    result = query(connection, tpl)

    if len(result) > 0 and getattr(result[0], version_column) is not None:
        return getattr(result[0], version_column) + 1
    else:
        return 1


class InputType(object):
    
    def __init__(self, name, params=None):
        if params is None:
            params = {}
  
        self.name = name
        self.params = params

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


def extract_number(number_str, decimal=None):
    #TODO: handle decimal=False/True
    #TODO: regexp \s*[-+]?\s*[€$%]?[0-9,\.]\+
    separators = []   
    groups = ['']
    groups_index = 0
    sign = None

    for c in number_str:

        if c in ['.', ',']:
            separators.append(c)
            groups.append('')
            groups_index += 1

        elif c in [' ', '€', '$', '%']:
            continue

        elif c in ['+', '-'] and sign is None:
            sign = c

        elif c in '1234567890':
            groups[groups_index] += c
            
            if sign is None:
                sign = '+'
        else:
            return None

    if len(groups) == 0:
        return None

    elif len(groups[0]) == 0:

        if len(''.join(groups)) == 0:
            return None

        number = Decimal('%s0.%s' % (sign, ''.join(groups)))
        precision = len(number.as_tuple().digits)
        scale = - number.as_tuple().exponent
        input_type = InputType('numeric', dict(precision=precision, scale=scale))
        return number, input_type

    elif len(groups) == 1:
        value = int(sign + groups[0])

        if value <= 2147483647 and value >= -2147483648:
            return value, InputType('int')

        elif value <= 9223372036854775807 and value >= -9223372036854775808:
            return value, InputType('bigint')

        else:
            return value, InputType('text')

    # xx.xx
    elif len(groups) == 2:
        number = Decimal('%s%s.%s' % (sign,''.join(groups[0:-1]), groups[-1]))
        precision = len(number.as_tuple().digits)
        scale = - number.as_tuple().exponent
        input_type = InputType('numeric', dict(precision=precision, scale=scale))
        return number, input_type

    # x.xxx,xx / x,xxx.xx
    elif len(groups) == 3 and separators[-1] != separators[-2]:
        number = Decimal('%s%s.%s' % (sign,''.join(groups[0:-1]), groups[-1]))
        precision = len(number.as_tuple().digits)
        scale = - number.as_tuple().exponent
        input_type = InputType('numeric', dict(precision=precision, scale=scale))
        return number, input_type


    else:
        return None


def detect_dayfirst(dates):

    first_test = None
    second_test = None

    try:
        for i in dates:
            dateutil_parse(i, dayfirst=True)

        first_test = True
    except:
        first_test = False

    try:
        for i in dates:
            dateutil_parse(i, dayfirst=False)

        second_test = True
    except:
        second_test = False
   
    if first_test != second_test:
        return first_test
    else:
         None


def parse_csv(input_data, delimiter=';', quotechar='"'):

    sniffer = csv.Sniffer()
    try:
        dialect = sniffer.sniff(input_data, delimiters=';\t,')
    except:
        csv.register_dialect('dlb_excel', delimiter=delimiter, quotechar=quotechar)
        dialect = csv.get_dialect('dlb_excel')

    data = input_data
    data = data.splitlines()
    data = csv.reader(data, dialect=dialect)

    # remove redundant new lines
    data = filter(lambda i: len(i) != 0, data)

    return list(data)


def detect_type(value, current_type=None):
    #TODO: Optimize, this is where 80-90% time is spent when loading data
    if value is None:
        return None, current_type       

    if current_type is None:
        extract_result = extract_number(value)

        if extract_result is not None:
            return extract_result

        try:
            result, dayfirst = dateutil_parse(value)
            #TODO: handle timezones
            return result, InputType('timestamp', dict(dayfirst=dayfirst, last_value=result))

        except Exception as e:
            pass
                    
        if value.lower() in ['true', 't', 'yes', 'y']:
            return True, InputType('bool')

        elif value.lower() in ['false', 'f', 'no', 'n']:
            return False, InputType('bool')

        return value, InputType('text')

    elif current_type.name == 'text':
        return str(value), InputType('text')

    elif current_type.name == 'numeric':

        if value == '' or value is None:
            return None, current_type

        extract_result = extract_number(value)

        # not a number
        if extract_result is None:
            return value, InputType('text')

        # it is a number
        new_type = extract_result[1]
        value = extract_result[0]

        if new_type.name == 'numeric':
            # enlarge precision to hold new value if necessary
            
            precision = max(current_type.params['precision'], new_type.params['precision'])
            scale = max(current_type.params['scale'], new_type.params['scale'])

            # make sure that it will hold old value
            natural_lenght = max(precision - scale, current_type.params['precision'] - current_type.params['scale'])
            precision = max(precision, natural_lenght+scale)
        
        elif new_type.name == 'int':
            natural_lenght = len(str(value))
            precision = max(current_type.params['precision'], natural_lenght)
            scale = current_type.params['scale']
            precision = max(natural_lenght + scale, precision)

        else:
            return value, InputType('text')
            
        return value, InputType('numeric', dict(precision=precision, scale=scale))

    elif current_type.name == 'int' or current_type.name == 'integer' or current_type.name == 'bigint':

        if value == '' or value is None:
            return None, current_type

        else:
            extract_result = extract_number(value)

            # not a number
            if extract_result is None:
                return value, InputType('text')

            return extract_result
        
    elif current_type.name == 'bool':
        if value.lower() in ['true', 't', 'yes', 'y']:
            return True, InputType('bool')

        elif value.lower() in ['false', 'f', 'no', 'n']:
            return False, InputType('bool')

        else:
            return None, current_type

    elif current_type.name.startswith('timestamp'):

        if value == '' or value is None:
            return None, current_type

        #TODO: handle timezones
        result = re.findall('\d+|jan|feb|mar|apr|jun|jul|aug|sep|oct|nov|dec', value.lower())
        if len(result) < 3 or len(result) > 8:
            return value, InputType('text')

        dayfirst = None

        try:
            if current_type.params.get('dayfirst') is None:
                #FIXME: dayfirst = detect_dayfirst(current_type.params['last_value'])
                #dayfirst = detect_dayfirst(value)
                result, dayfirst = dateutil_parse(value)
            else:
                dayfirst = current_type.params['dayfirst']
                result, dayfirst = dateutil_parse(value, dayfirst=dayfirst)

        except Exception as e:
            return value, InputType('text')

        return result, InputType('timestamp', dict(dayfirst=dayfirst, last_value=value))

    else:
        return value, current_type


def preprocess_csv_data(input_data,):
    data = parse_csv(input_data)
    
    if len(data) < 2:
        raise ValueError('Not enough data')

    header = data[0]
    formated_header = format_header(header)
    
    return formated_header, header, data[1:]


def get_row_data_types(header, row, current_types=None, detect_type_func=None):

    if detect_type_func is None:
        detect_type_func = detect_type

    if current_types is None:
        current_types = {}

    new_types = dict(current_types)

    result_row = []

    for i, value in enumerate(row):

        if i >= len(header):
            break

        v, new_types[header[i]] = detect_type_func(value, current_types.get(header[i]))
        result_row.append(v)

    return result_row, new_types


def get_data_types(header, data, current_types=None, detect_type_func=None):

    if detect_type_func is None:
        detect_type_func = detect_type

    if current_types is None:
        current_types = {}

    result_data = []
    for row in data:

        result_row = []
        empty_count = 0

        for value in row:
            if value == '':
                empty_count +=1

        if empty_count == len(row):
            continue

        for i, value in enumerate(row):
            v, current_types[header[i]] = detect_type_func(value, current_types.get(header[i]))

            result_row.append(v)

        result_data.append(result_row)


    for h in header:
        if h not in current_types:
            current_types[h] = 'text'
        

    return result_data, current_types


def format_header(header):

    formated_header = []
    collisions = {}
    generic_name_counter = 0

    for column in header:
        column = unidecode.unidecode(column)
        letters_and_digits = string.ascii_lowercase + string.digits
        buf = ''

        if len(column) > 0:
            for i, char in enumerate(column.lower()):

                if i == 0 and char in string.digits:
                    buf += '_'

                elif char not in letters_and_digits:
                    buf += '_'

                else:
                    buf += char

            # check for collisions
            if buf in formated_header:
                suffix = collisions.get(buf)

                if suffix is not None:
                    collisions[buf] = suffix+1 
                    buf = buf + '_' +  str(suffix+1)

                else:
                    buf += '_1'
        else:
            generic_name_counter += 1
            buf = 'column_%i' % generic_name_counter

        formated_header.append(buf)

    return formated_header


if __name__ == '__main__':
    print(detect_type('2017-05-17 12:27:25.2945'))



import itertools
import os.path
from operator import itemgetter
from threading import Thread
from threading import Semaphore
from itertools import islice, zip_longest
import traceback

try:
    from queue import Queue
except ImportError:
    from Queue import Queue


from witchcraft.utils import build_tuple_type, __query_paths
from witchcraft.utils import coalesce, chainlist
from witchcraft.template import template

__buildin_filter = filter
__buildin_map = map


base_path = os.path.dirname(os.path.abspath(__file__))


def th(*operations):
    args = []
    for item in operations:
        args.extend(list(item)[1:])
        args = [item[0](*args)]

    return args[0]


def set_query_path(path):
    global __query_paths

    if isinstance(path, list):
        __query_paths = path.extend(__query_paths)

    elif isinstance(path, str) or isinstance(path, unicode):
        __query_paths.insert(0, path)

    __query_paths.append(os.path.join(base_path, 'queries'))


def dict_merge(a, b):
    c = b.copy()
    c.update(a)
    return c
  

#TODO: think about how to make query lazy
def query(connection, sql_query):
    if connection.connection is not None and callable(getattr(connection.connection, 'execute', None)):
        result_proxy = connection.connection.execute(sql_query)
    else:
        result_proxy = connection.execute(sql_query)

    result_type = build_tuple_type(*result_proxy.keys())
    result = list(map(lambda r: result_type(dict(r)), result_proxy))
    result_proxy.close()
    return result
 

def batch_fetch(connection, sql_query, batch_size):
    if connection.connection is not None and callable(getattr(connection.connection, 'execute', None)):
        result_proxy = connection.connection.execute(sql_query)
    else:
        result_proxy = connection.execute(sql_query)

    if result_proxy.returns_rows:
        result_type = build_tuple_type(*result_proxy.keys())
        result = map(lambda r: result_type(dict(r)), result_proxy)

        while True:
            chunk = list(islice(result, batch_size))

            if len(chunk) != 0:
                yield chunk
            else:
                break

    result_proxy.close()


def execute(connection, sql_query):
    if connection.connection is not None and callable(getattr(connection.connection, 'execute', None)):
        result_proxy = connection.connection.execute(sql_query)
    else:
        result_proxy = connection.execute(sql_query)

    row_count = result_proxy.rowcount
    result_proxy.close()
    return row_count


def filter(data, func):
    return __buildin_filter(func, data) 


def each(data, func):
    return list(__buildin_map(func, data))


def filter_by(data, key_name, value):
    
    def filter_fn(item):
        return item[key_name] == value
    
    return __buildin_filter(filter_fn, data) 


def filter_in(data, key_name, values):

    def filter_fn(item):
        return item[key_name] in values

    return __buildin_filter(filter_fn, data)


def group_by(data, key_func, *operations):
    memo = {}

    for item in data:
        key = key_func(item)
        group_list = memo.get(key, [])
        group_list.append(item)
        memo[key] = group_list

    for key, value in memo.items():

        for op_fn in operations:

            if isinstance(op_fn, tuple):
                args = [value]
                args.extend(list(op_fn)[1:])
                value = op_fn[0](*args)

            else:
                value = op_fn(value)

        memo[key] = value

    return memo


def group_by_columns(data, key_names, *operations):
    operations = list(operations)

    def post_process(d):
        return ommit_columns(d, key_names)

    operations.insert(0, post_process)
    return group_by(data, lambda d: select_columns(d, key_names), *operations)


def columns(data, columns):
    return to_tuple(data, columns)


def select_columns(data, columns):
    dest_column_names = []
    src_column_names = []    

    for c in columns:
        
        if isinstance(c, list) or isinstance(c, tuple):
            src_column_names.append(c[0])
            dest_column_names.append(c[1])
        else:
            src_column_names.append(c)
            dest_column_names.append(c)

    tuple_type = build_tuple_type(*dest_column_names)

    if isinstance(data, list):
        return list(map(lambda i: tuple_type(i, src_column_names), data))
    else:
        return tuple_type(data, src_column_names)


def ommit_columns(tuple_set, columns):
    columns = set(columns)

    def ommit_fn(item):
        keys = item.keys()
        keys = list(set(item.keys()) - set(columns))
        result_type = build_tuple_type(*keys)

        if isinstance(item, dict):
            return result_type({k:item[k] for k in item if k in keys})
        else:
            return result_type(item.select(keys))
        
    return list(map(ommit_fn, tuple_set))


def add_column_with(tuple_set, column_name, func):

    def add_fn(item):

        result_type = build_tuple_type(*(list(item.keys()) + [column_name]))

        if isinstance(item, dict):
            r = dict(**item)
        else:
            r = item.asdict()

        r[column_name] = func(r)
        return result_type(r)

    return list(map(add_fn, tuple_set))


def add_column(tuple_set, column_name, value):
    return add_column_with(tuple_set, column_name, lambda r: value)


def set_column_with(tuple_set, column_name, func):

    def set_fn(item):
        result_type = build_tuple_type(*list(item.keys()))

        if isinstance(item, dict):
            r = dict(**item)
        else:
            r = item.asdict()

        r[column_name] = func(r)
        return r

    return list(map(set_fn, tuple_set))


def set_column(tuple_set, column_name, value):
    return set_column_with(tuple_set, column_name, lambda r: value)


def aggregate(iterable, *aggregators):
    memo = iterable

    for agg_fn in aggregators:
        memo = agg_fn(iterable, memo)

    return memo


def aggregate_list_of(iterable, extract_key):
    result = list(map(lambda i: i[extract_key], iterable))
    return result

def aggregate_lists_of(iterable, extract_key):
    result = list(map(lambda i: i[extract_key], iterable))
    return list(itertools.chain(*result))

def aggregate_first_row(iterable):
    return iterable[0]


def aggregate_first_of(iterable, extract_key):
    return list(iterable)[0][extract_key]


def _flatten(keys, items, column_names):
    #if (len(column_names) - len(keys)) == 0:
    #    return items
    key_names = column_names[:len(keys)]
    result = []
    key_part = dict(zip(key_names, keys))
    if isinstance(items, dict):
        first_values = items.values()

        #print 'first_values', first_values
        #if len(first_values) and (isinstance(first_values, dict) or isinstance(first_values, OrderedDict)):
        #    items = flatten_all(items, column_names[len(keys):])
        #else:
        #print('first_values',first_values)
        #if not (isinstance(first_values, tuple) or isinstance(first_values, list)):
        #    items = (items, )

        if len(column_names[len(keys):]):
            #print items, column_names
            items = flatten_dict(items, column_names[len(keys):])

        else:
            items = (items, )

    if not (isinstance(items, tuple) or isinstance(items, list)):
        items = (items, )
    
    for i in items:

        # can be converted to dictionary
        asdict_op = getattr(i, "asdict", None)
        if callable(asdict_op):
            i = i.asdict()
        
        # is dictionary
        elif isinstance(i, dict):
            pass

        #is value
        else:
            i = {'value': i}

        result.append(dict(key_part, **i))

    return result
    

#TODO: really required??? - No, itemize_dict is sufficient
def flatten_dict(data, column_names):
    result = []

    if isinstance(data, dict):
        #print data.items()
        for key, i in data.items():

            if len(i) == 0:
                continue

            #print key
            if not (isinstance(key, tuple) or isinstance(key, list)):
                key = (key,)
            else:
                key = key.values()

            result.extend(_flatten(key, i, column_names))

        return to_tuple(result, column_names)

    else:
        return to_tuple(data, column_names)


def flatten(mapping, default_keys=None):

    keys = list(mapping.keys())
    values = mapping.values()

    def find_non_empty(v):
        for i in v:
            if len(i) > 0:
                return i

    if len(keys) == 0:
        return []

    value = find_non_empty(values)
    
    if value is None:
        return []

    elif isinstance(value, list):
        list_value = True
        value_keys = list(value[0].keys())

    else:
        list_value = False
        value_keys = list(value.keys())

    keys = keys[0].keys() + value_keys

    if default_keys is None:
        ResultType = build_tuple_type(*keys)
    else:
        ResultType = build_tuple_type(*default_keys)

    result = []

    if list_value:

        for key, value in mapping.items():

            for item in value:
                result.append(ResultType(key.items() + list(item.items())))
                
    else:
        for key, value in mapping.items():
            result.append(ResultType(key.items() + value.items()))

    return result


# itemize(['name'] 'columns')
def itemize_dict(mapping, columns):
    result = []
    result_type = build_tuple_type(*columns)

    for key, value in mapping.items():
        item = {}

        # key is not dictionary
        if not callable(getattr(key,'values', None)):
            key = {'value': key}

        for i, c in enumerate(columns[:-1]):
            item[c] = list(key.values())[i]

        item[columns[-1]] = value
        result.append(result_type(item))

    return result


def complement(left, right, combine_fn=None):
    left_keys = set(left.keys())
    right_keys = set(right.keys())

    intersection = left_keys - right_keys
    
    result = {}

    if combine_fn is None:
        for key in intersection:
                result[key] = left[key]
    
    else:
         for key in intersection:
                result[key] = combine_fn(left[key])
       
    return result


def intersect(left, right, combine_fn=None):
    left_keys = set(left.keys())

    if isinstance(right, dict):
        right_keys = set(right.keys())

    else:
        right_keys = right

    intersection = left_keys & right_keys
    
    result = {}

    if combine_fn is not None:

        if isinstance(right, dict):

            for key in intersection:
                result[key] = combine_fn(left[key], right[key], key)

        else:

            for key in intersection:
                result[key] = combine_fn(left[key], key)

    else:

        for key in intersection:
            result[key] = left[key]

    return result


def join_by_columns(left, right, left_keys, right_keys):

    if len(left) == 0:
        return []

    mlk = set(left[0].keys())
    mrk = set(right[0].keys()) - set(right_keys)

    tuple_type = build_tuple_type(list(mrk) + list(mlk))

    def merge(left, right, k):
        result = []

        if len(right) == 0:

            for li in left:
                r = tuple_type(li.items() + k.items())
                result.append(r)

        else:
            for li in left:
                
                for ri in right:
                    r = tuple_type(dict(ri.items() + li.items() + k.items()))
                    result.append(r)
            
        return result

    left = group_by_columns(left, left_keys)
    right = group_by_columns(right, right_keys)

    for k,lv in left.items():
        rv = right.get(k)

        if rv is not None:
           left[k] = merge(lv, rv, k)
 
    return flatten(left, list(mrk)+list(mlk))


#TODO: join_by_fn
def union(iterables):
    return itertools.chain.from_iterable(iterables)


def to_tuple(data, keys=None):
    if keys is None:
        keys = data[0].keys()
    tuple_type = build_tuple_type(*keys)
    return list(map(tuple_type, data))


def sort_by_columns(iterable, *columns):
    return sorted(iterable, key=itemgetter(*columns))


def in_list(search_list):
    return lambda v: v in search_list


def parallel_map(func, maxthreads, items):
    semaphore = Semaphore(0)

    def worker():

        while True:
            item = q.get()

            try:
                import os
                func(item)

            except:
                print(traceback.format_exc())
    
            q.task_done()
            semaphore.release()

    q = Queue(maxthreads)

    for i in range(maxthreads):
         t = Thread(target=worker)
         t.daemon = True
         t.start()

    for c,item in enumerate(iter(items)):

        if c > 10000000:
            print("Safety stop condition reached")
            break

        q.put(item)

    q.join()


def distinct(iterable, *columns):

    keys = set()    

    def ffn(item):
        key = set()

        for cn in columns:
            cv = item.get(cn, None)

            if cv is not None:
                key.add(cv)

        key = tuple(key)

        if key not in keys:
            keys.add(key)
            return True
        
        return False

    return __buildin_filter(ffn, iterable)


def apply_func(func, args, kwargs):
    return func(*args, **kwargs)


def equal(left, right):
    
    if isinstance(left, list) and isinstance(right, list):

        for l,r in zip_longest(left,right):
            if l != r:
                return False

        return True

    elif not isinstance(left, list) and not isinstance(right, list):
        return left == right

    else:
        raise ValueError("Both arguments must be eigher of type 'list' or ")




#!/usr/bin/env python
from setuptools import setup

setup(
    name="witchcraft",
    version="0.2.54",
    description='',
    author='Peter Facka',
    author_email='pfacka@spookyowl.com',
    url='https://github.com/spookyowl/witchcraft',
    packages=[
        'witchcraft',
        'witchcraft.dateutil',
        'witchcraft.dateutil.tz'
    ],
    zip_safe=False,	
    install_requires=[
	'psycopg2>=2.6.1',
	'SQLAlchemy>=1.0.6',
        'hy>=0.11.1',
        'pyparsing>=2.1.1',
        'unidecode'
    ],
    provides=['witchcraft'],
    include_package_data=True,
    classifiers=[
      'Development Status :: 3 - Alpha',
      'Environment :: Web Environment',
      'Intended Audience :: Developers',
      'Operating System :: Microsoft :: Windows',
      'Operating System :: MacOS :: MacOS X',
      'Operating System :: POSIX',
      'Programming Language :: Python :: 2.7',
      'Programming Language :: Python :: 3.3',
      'License :: OSI Approved :: MIT License',
      ],
)



import psycopg2
import psycopg2.extras

connection = psycopg2.connect(
    f'postgresql://user:password@postgresql/testing_db',
    connection_factory=psycopg2.extras.LogicalReplicationConnection
)

cursor = connection.cursor()
replication_slot_name = 'testing'

try:
    cursor.create_replication_slot(replication_slot_name, output_plugin='wal2json')
except Exception:
    pass

cursor.start_replication(slot_name=replication_slot_name, decode=True, status_interval=1)
cursor.consume_stream(lambda message: print('received message:', message, message.payload))



import sqlalchemy


def get_value(data, keys, default=None):
    result = None

    for k in keys:
        result = data.get(k)
        
        if result is not None:
            return result

    if result is None:
        return default


class Connection(object):
    
    def __init__(self, auth):
        url = self._build_url(auth)
        self.database_type = get_value(auth, ['type', 'db_type'], 'pgsql')
        self.engine = sqlalchemy.create_engine(url, convert_unicode=True)
        self.connection = None
        self.transaction = None

    def connect(self):
        self.connection = self.engine.connect()

    def disconnect(self):
        self.connection.close()
        self.connection = None

    def execute(self, sql_query):
        if not self.is_connected():
            self.connect()

        return self.connection.execute(sql_query)

    def begin_transaction(self):
        if not self.is_connected():
            self.connect()
        self.transaction = self.connection.begin()

    def end_transaction(self):
        self.transaction.commit()
        self.transaction = None

    def is_connected(self):
        return self.connection is not None

    def _build_url(self, auth):


        db_type = get_value(auth, ['type', 'db_type'], 'pgsql')

        if db_type == 'mysql':
            default_port = 3306

        elif db_type == 'pgsql':
            default_port = 5432

        ctx = (get_value(auth, ['user']),
               get_value(auth, ['passwd', 'password', 'pass']),
               get_value(auth, ['host', 'server'], 'localhost'),
               get_value(auth, ['port'], default_port),
               get_value(auth, ['database', 'db_name', 'database_name', 'db']))

        if db_type == 'pgsql':
            url_tpl = 'postgresql+psycopg2://%s:%s@%s:%s/%s' % ctx

        elif db_type == 'mysql':
            url_tpl = 'mysql+mysqldb://%s:%s@%s:%s/%s' % ctx

        else:
            RaiseValue('db_type must be eighter "mysql" or "pgsql"')

        return url_tpl % auth


class Session(object):

    def __init__(self, args):
        self.args = args

    def __enter__(self):
        self.connection = Connection(self.args)
        self.connection.connect()
        return self.connection

    def __exit__(self, type, value, traceback):
        self.connection.disconnect()



